import click

from pmx_glossary.glossary import Glossary


@click.group()
def cli():
    Glossary.initialize()


@cli.command()
def jsondump():
    """
    Dump all entry data to a json file.
    """
    click.secho('Dumping to json...')
    data, filepath = Glossary.store.dump_to_json()

    entry_count = len(data['entries'])
    redirect_count = len(data['redirects'])

    print(
        'Dumped {} glossary entries and {} redirects to {}'.format(
            entry_count, redirect_count, filepath
        )
    )


@cli.command()
@click.argument('path', type=click.Path(exists=True))
def jsonload(path):
    """
    Load entry data from a json file.
    """
    all_data, inserted = Glossary.store.load_from_json(path)
    all_entries_count = len(all_data['entries'])
    inserted_entries_count = len(inserted)

    print(
        'Inserted {}/{} entries from file.'.format(
            inserted_entries_count, all_entries_count
        )
    )

    if all_entries_count != inserted_entries_count:
        print(
            'This implies that {} entries were already in the database.'.format(
                all_entries_count - inserted_entries_count
            )
        )

    print


@cli.command()
@click.argument('path', type=click.Path(exists=True))
def load_fixtures(path):
    Glossary.load_fixtures(path)



import os
import datetime
import random
import unittest

from pmx_glossary import glossary


class GlossaryTestCase(unittest.TestCase):
    DB_FILE = 'pmxbot_test.sqlite'

    TEST_DEFINITIONS = {
        'blargh': 'this one thing I had',
        'building': 'where salmon hold meetings',
        'castle': 'where salmon have tea',
        'fish Oil': 'what salmon sell',
        'Salmon-person': 'a type of things',
        'snargh': 'also a thing',
        'snowman': u'\u2603',
        '__underscores__': 'woah',
    }

    TEST_NICK = 'tester_person'

    def setUp(self):
        self.wipe_and_init_glossary()
        self.store = glossary.Glossary.store

    def wipe_and_init_glossary(self):
        if os.path.exists(self.DB_FILE):
            os.remove(self.DB_FILE)

        glossary.Glossary.initialize(
            'sqlite:' + self.DB_FILE, load_fixtures=False
        )

    def tearDown(self):
        glossary.pmxbot.storage.SelectableStorage.finalize()

    def _load_test_definitions(self, definition_dict=None):
        definition_dict = definition_dict or self.TEST_DEFINITIONS

        for entry, definition in definition_dict.items():
            self._call_define(
                u'{}: {}'.format(entry, definition), nick=self.TEST_NICK
            )

    def _call_command(self, func, rest, nick=None):
        nick = nick or self.TEST_NICK

        return func(
            client='client',
            event='event',
            channel='channel',
            nick=nick,
            rest=rest
        )

    def _call_search(self, rest):
        return glossary.search(
            client='client',
            event='event',
            channel='channel',
            nick=self.TEST_NICK,
            rest=rest
        )

    def _call_define(self, rest, nick=None):
        nick = nick or self.TEST_NICK

        return glossary.define(
            client='client',
            event='event',
            channel='channel',
            nick=nick,
            rest=rest
        )

    def _call_whatis(self, rest, nick=None):
        nick = nick or self.TEST_NICK

        return glossary.query_command(
            client='client',
            event='event',
            channel='channel',
            nick=nick,
            rest=rest
        )

    def _call_redirect(self, rest, nick=None):
        nick = nick or self.TEST_NICK

        return glossary.redirect_command(
            client='client',
            event='event',
            channel='channel',
            nick=nick,
            rest=rest
        )

    def _call_unredirect(self, rest, nick=None):
        return self._call_command(glossary.remove_redirect, rest, nick)

    def _call_whowrote(self, rest, nick=None):
        return self._call_command(glossary.who_wrote, rest, nick)

    def test_dump_and_load(self):
        self._load_test_definitions()

        redirects = 'red1', 'red2', 'red3', 'red4'
        redirect_map = {}

        for redirect_from in redirects:
            redirect_to = random.choice(self.TEST_DEFINITIONS.keys())
            redirect_map[redirect_from] = redirect_to

            self._call_redirect('{}: {}'.format(redirect_from, redirect_to))

        self.assertEqual(len(redirects), len(redirect_map.keys()))

        # Dump the data
        dump_data, filepath = glossary.Glossary.store.dump_to_json()

        dumped_entries = {r['entry'] for r in dump_data['entries']}
        expected_entries = set(self.TEST_DEFINITIONS.keys())

        self.assertEqual(dumped_entries, expected_entries)

        # Reloading what's already in the db should not affect anything.
        loaded_json, inserted = glossary.Glossary.store.load_from_json(
            filepath
        )
        self.assertEqual(len(inserted), 0)
        self.assertEqual(len(loaded_json['entries']), len(expected_entries))

        # Loading into a clean db should insert all the things.
        self.wipe_and_init_glossary()
        glossary.Glossary.store.bust_all_entries_cache()

        loaded_entries, inserted = glossary.Glossary.store.load_from_json(
            filepath
        )
        self.assertEqual(len(inserted), len(expected_entries))

        # Did we get all our entries back?
        for entry in expected_entries:
            records = glossary.Glossary.store.get_all_records_for_entry(entry)

            self.assertEqual(1, len(records))

            self.assertEqual(records[0].entry, entry)

        # Did we get all our redirects back?
        for redirect_from, redirect_to in redirect_map.items():
            record = self.store.get_redirect(redirect_from)
            self.assertEqual(record.entry, redirect_to)

        os.remove(filepath)

    def test_add_and_retrieve_simple_definition(self):
        author = 'bojangles'
        entry = 'fish'
        definition = 'a swimmy thingy'
        rest = '{}: {}'.format(entry, definition)

        add_result = self._call_define(rest, nick=author)

        self.assertEqual(
            add_result,
            glossary.ADD_DEFINITION_RESULT_TEMPLATE.format(
                entry=entry,
                definition=definition,
                nth=glossary.nth_str(1)
            )
        )

        result = self._call_whatis(entry)
        expected = 'fish (1/1): a swimmy thingy [just now]'

        self.assertEqual(result, expected)

    def test_suggestions_on_query_miss(self):
        self._load_test_definitions({
            'shindig': 'a thing',
            'shining': 'shiny thing',
            'flabbergasted': 'wat!',
            'ning-ning-ning': 'no idea',
        })

        expected = (
            '"shin" is undefined. May I interest you in shindig or shining?'
        )
        result = self._call_whatis('shin')

        self.assertEqual(result, expected)

    def test_get_all_records(self):
        """
        All entries gets only the latest entry for a given entry_lower.
        """
        self._load_test_definitions({
            'fish': 'swimmer',
        })
        self._load_test_definitions({
            'FISH': 'big swimmer',
        })
        self._load_test_definitions({
            'fisH': 'awkward swimmer',
        })
        self._load_test_definitions({
            'FisH': 'symmetrical swimmer',
        })
        self._load_test_definitions({
            'onion': 'yumm',
        })
        self._load_test_definitions({
            'oNion': 'a thing',
        })
        self._load_test_definitions({
            'onioN': 'snappy',
        })

        records = glossary.Glossary.store.get_all_records()

        entries = {r.entry for r in records}

        self.assertEqual(entries, {'FisH', 'onioN'})

    def test_add_and_retrieve_entry_with_unicode(self):
        entry = u'\u2603'
        definition = u'a snowman, like \u2603'

        self._call_define(u'{}: {}'.format(entry, definition))

        result = self._call_whatis(entry)

        expected = glossary.QUERY_RESULT_TEMPLATE.format(
            entry=entry,
            num=1,
            total=1,
            definition=definition,
            author=self.TEST_NICK,
            age=glossary.datetime_to_age_str(datetime.datetime.utcnow()),
            channel_str=' in channel'
        )

        self.assertEqual(result, expected)

    def test_add_and_retrieve_entry_with_spaces(self):
        entry = 'a fish called wanda'
        definition = 'a fish named wanda'
        self._call_define('{}:{}'.format(entry, definition))

        result = self._call_whatis(entry)

        expected = glossary.QUERY_RESULT_TEMPLATE.format(
            entry=entry,
            num=1,
            total=1,
            definition=definition,
            author=self.TEST_NICK,
            age=glossary.datetime_to_age_str(datetime.datetime.utcnow()),
            channel_str=' in channel'
        )

        self.assertEqual(result, expected)

    def test_add_and_retrieve_multiple_definitions(self):
        author = 'bojangles'
        entry = 'fish'
        definitions = (
            'a swimmy thingy',
            'a swimmy slimy thingy',
            'dinner'
        )

        for i, definition in enumerate(definitions):
            result = self._call_define(
                '{}: {}'.format(entry, definition), nick=author
            )

            self.assertIn(
                'This is the {} time it has been defined'.format(
                    glossary.nth_str(i + 1)
                ),
                result
            )

        expected_zero = '"fish" has 3 records. Valid record numbers are 1-3.'
        self.assertEqual(self._call_whatis('fish: 0'), expected_zero)

        # Fetch the first definition
        expected_1 = "fish (1/3): a swimmy thingy [just now]"
        result = self._call_whatis('fish: 1')
        self.assertEqual(result, expected_1)

        # Fetch the second definition
        expected_2 = "fish (2/3): a swimmy slimy thingy [just now]"
        result = self._call_whatis('fish: 2')
        self.assertEqual(result, expected_2)

        # Fetch the third definition
        expected_3 = "fish (3/3): dinner [just now]"
        result = self._call_whatis('fish: 3')
        self.assertEqual(result, expected_3)

        # Fetch the default (latest) definition
        result = self._call_whatis('fish')
        self.assertEqual(result, expected_3)

    def test_get_random_definition(self):
        self._load_test_definitions()

        expected_entries = set(self.TEST_DEFINITIONS.keys())

        for i in range(20):
            result = self._call_whatis('')
            entry = result.split('(', 1)[0].strip()
            definition = result.split(':')[-1].strip()

            self.assertIn(entry, expected_entries)

            expected_definition = (
                u'{} [just now]'.format(self.TEST_DEFINITIONS[entry])
            )
            self.assertEqual(definition, expected_definition)

    def test_all_entries(self):
        self._load_test_definitions()

        all_records = set(glossary.Glossary.store.get_all_records())
        all_entries = {r.entry for r in all_records}

        self.assertEqual(all_entries, set(self.TEST_DEFINITIONS.keys()))

        # Make sure cache is invalidated by new addition
        new_entry = 'new_entry'

        glossary.Glossary.store.add_entry(
            entry=new_entry,
            definition='new def',
            author='author'
        )

        all_records_again = glossary.Glossary.store.get_all_records()

        self.assertEqual(
            all_entries | {new_entry, },
            {r.entry for r in all_records_again}
        )

    def test_get_words_like(self):
        entry_dict = {
            'a fish': 'just a fish',
            'fishy': 'fishlike',
            'gone fishing': 'fishin',
            'gofish': 'sweet game',
        }

        self._load_test_definitions(entry_dict)

        result = set(glossary.Glossary.store.get_similar_words('fish'))

        self.assertEqual(result, set(entry_dict.keys()))

    def test_punctuation_error_response(self):
        for punct_char in glossary.INVALID_ENTRY_CHARS:
            entry = 'entry' + punct_char
            definition = 'a super disallowed entry'

            result = self._call_define('{}: {}'.format(entry, definition))

            if punct_char == ':':
                expected = (
                    "I can't handle '::' right now. "
                    "Please try again without it."
                )
                self.assertEqual(expected, result)
            else:
                expected = (
                    '"{}" cannot be used in a glossary entry.'
                ).format(punct_char)

                self.assertEqual(result, expected)

    def test_get_alternative_suggestions(self):
        self._load_test_definitions({
            'dataplasm': 'a plasm full of data',
            'octonerd': 'eight nerds, enmeshed',
            'terse herd': 'a concise form of herd'
        })

        self.assertEqual(
            glossary.get_alternative_suggestions('terse data'),
            {'terse herd', 'dataplasm'}
        )

        self.assertEqual(
            glossary.get_alternative_suggestions('octo-plasm'),
            {'octonerd', 'dataplasm'}
        )

        self.assertEqual(
            glossary.get_alternative_suggestions('plasm_herd'),
            {'terse herd', 'dataplasm'}
        )

        self.assertEqual(
            glossary.get_alternative_suggestions('zamboni'),
            set()
        )

    def test_search(self):
        self._call_define('fish: swimmer')
        self._call_define('fisH: shugga')
        self._call_define('fish head: the top part of a fish')
        self._call_define('fissure: rip')
        self._call_define('zamboni: big thingie')
        self._call_define('FisH: shugga bomb')

        result = self._call_search('fish')

        self.assertIn(': FisH and fish head.', result)

    def test_simple_redirect(self):
        self._call_define('Bad Dude: one bad summagun')
        self._call_define('cool Dude: one cool summagun')

        self._call_redirect('bad Dude: cool DUde')

        redirect = self.store.get_redirect('bad dude')

        self.assertEqual(redirect.entry, 'cool Dude')

        result = self._call_whatis('bad dude')
        expected = glossary.QUERY_RESULT_TEMPLATE.format(
            entry='cool Dude',
            num=1,
            total=1,
            definition='one cool summagun',
            age='just now'
        )

        # Capitalization of bad dude should reflect what user passed in,
        # not what's stored in entry.
        expected = 'bad dude redirects to ' + expected

        self.assertEqual(result, expected)

    def test_redefine_redirect(self):
        self._load_test_definitions({
            'thingy': 'flamma jam',
            'tom tom tom': 'superdrum',
        })

        for _ in range(5):
            self.store.add_redirect('thingy', 'tom tom tom')
            self.assertEqual(
                self.store.get_redirect('thingy').entry, 'tom tom tom'
            )

    def test_redirect_to_undefined(self):
        result = self._call_redirect('ting: thing')
        self.assertEqual(result, '"thing" is not defined')

    def test_circular_redirect(self):
        self._load_test_definitions({
            'thingy': 'flamma jam',
            'tom tom tom': 'superdrum',
        })

        self._call_redirect('thingy: tom tom tom')
        r_record = self.store.get_redirect('thingy')
        self.assertEqual(r_record.entry, 'tom tom tom')

        invalid_attempt = self._call_redirect('tom tom tom: thingy')
        self.assertEqual(
            invalid_attempt,
            '"thingy" is itself being redirected to "tom tom tom."'
        )

    def test_remove_redirect(self):
        self._load_test_definitions({
            'thingy': 'flamma jam',
            'tom tom tom': 'superdrum',
        })

        self._call_redirect('thingy: tom tom tom')
        self.assertEqual(self.store.get_redirect('thingy').entry, 'tom tom tom')

        self._call_unredirect('thingy')
        self.assertIsNone(self.store.get_redirect('thingy'))

        self.assertEqual(
            self._call_whatis('thingy'),
            glossary.QUERY_RESULT_TEMPLATE.format(
                entry='thingy',
                num=1,
                total=1,
                definition='flamma jam',
                author=self.TEST_NICK,
                age='just now'
            )
        )

    def test_who_wrote(self):
        nick1 = 'happyfeller'
        nick2 = 'happyfeller_blues'
        self._call_define('thing: a thing, you know?', nick=nick1)
        self._call_define('thing: thing thing thing', nick=nick2)

        expected = nick2 + ' authored the 2nd definition of thing.'
        result = self._call_whowrote('thing')
        self.assertEqual(result, expected)

        expected_1 = nick1 + ' authored the 1st definition of thing.'
        result_1 = self._call_whowrote('thing: 1')
        self.assertEqual(result_1, expected_1)

    def test_who_wrote_with_redirect(self):
        nick = 'happyfeller'
        self._call_define('thing: a thing, you know?', nick=nick)
        self._call_redirect('ting: thing', nick=nick)

        expected = (
            'ting redirects to thing. {} authored the 1st definition of thing.'
        ).format(nick)
        result = self._call_whowrote('ting')

        self.assertEqual(result, expected)


class ReadableJoinTestCase(unittest.TestCase):
    def test_no_items(self):
        self.assertEqual(None, glossary.readable_join([]))

    def test_one_item(self):
        self.assertEqual('thing', glossary.readable_join(['thing']))

    def test_two_items(self):
        self.assertEqual(
            'thing1 or thing2', glossary.readable_join(['thing1', 'thing2'])
        )

    def test_three_items(self):
        self.assertEqual(
            'thing1, thing2, or thing3',
            glossary.readable_join(['thing1', 'thing2', 'thing3'])
        )

    def test_four_items(self):
        self.assertEqual(
            'thing1, thing2, thing3, or thing4',
            glossary.readable_join(['thing1', 'thing2', 'thing3', 'thing4'])
        )


class AgeStringTestCase(unittest.TestCase):
    @property
    def now(self):
        return datetime.datetime.utcnow()

    def test_just_now_str(self):
        self.assertEqual('just now', glossary.datetime_to_age_str(self.now))

    def test_just_now_str_under_minute(self):
        dt = self.now - datetime.timedelta(seconds=55)
        self.assertEqual('just now', glossary.datetime_to_age_str(dt))

    def test_minute_ago(self):
        dt = self.now - datetime.timedelta(seconds=60)
        self.assertEqual('1 minute ago', glossary.datetime_to_age_str(dt))

    def test_minutes_ago(self):
        dt = self.now - datetime.timedelta(seconds=60 * 55)
        self.assertEqual('55 minutes ago', glossary.datetime_to_age_str(dt))

    def test_hour_ago(self):
        dt = self.now - datetime.timedelta(seconds=60 * 60)
        self.assertEqual('1 hour ago', glossary.datetime_to_age_str(dt))

    def test_hours_ago_str(self):
        dt = self.now - datetime.timedelta(hours=13)
        self.assertEqual('13 hours ago', glossary.datetime_to_age_str(dt))

    def test_yesterday_str(self):
        dt = self.now - datetime.timedelta(days=1)
        self.assertEqual('yesterday', glossary.datetime_to_age_str(dt))

    def test_two_days_ago(self):
        dt = self.now - datetime.timedelta(days=2)
        self.assertEqual('2 days ago', glossary.datetime_to_age_str(dt))

    def test_30_days_ago(self):
        dt = self.now - datetime.timedelta(days=30)
        self.assertEqual('30 days ago', glossary.datetime_to_age_str(dt))

    def test_31_days_ago(self):
        dt = self.now - datetime.timedelta(days=31)
        self.assertEqual('1.0 months ago', glossary.datetime_to_age_str(dt))

    def test_40_days_ago(self):
        dt = self.now - datetime.timedelta(days=40)
        self.assertEqual('1.3 months ago', glossary.datetime_to_age_str(dt))

    def test_100_days_ago(self):
        dt = self.now - datetime.timedelta(days=100)
        self.assertEqual('3.3 months ago', glossary.datetime_to_age_str(dt))

    def test_365_days_ago(self):
        dt = self.now - datetime.timedelta(days=365)
        self.assertEqual('1.0 years ago', glossary.datetime_to_age_str(dt))

    def test_450_days_ago(self):
        dt = self.now - datetime.timedelta(days=450)
        self.assertEqual('1.2 years ago', glossary.datetime_to_age_str(dt))


class NthStringTestCase(unittest.TestCase):
    def test_nth_string(self):
        self.assertEqual(glossary.nth_str(1), '1st')
        self.assertEqual(glossary.nth_str(2), '2nd')
        self.assertEqual(glossary.nth_str(3), '3rd')
        self.assertEqual(glossary.nth_str(4), '4th')
        self.assertEqual(glossary.nth_str(5), '5th')
        self.assertEqual(glossary.nth_str(1000), '1000th')



import calendar
import datetime
import json
import random
import string
import tempfile
from collections import namedtuple

import pmxbot
from pmxbot import storage
from pmxbot.core import command, AliasHandler, CommandHandler

import pmx_glossary

DEFINE_COMMAND = 'define'
QUERY_COMMAND = 'whatis'
SEARCH_COMMAND = 'search'
WHOWROTE_COMMAND = 'whowrote'
REDIRECT_COMMAND = 'redirect'
REMOVE_REDIRECT_COMMAND = 'unredirect'
ARCHIVES_LINK_COMMAND = 'tardis'
ADMIN_COMMAND = 'glos'

HELP_DEFINE_STR = '!{} <entry>: <definition>'.format(DEFINE_COMMAND)
HELP_QUERY_STR = '!{} <entry> [: num]'.format(QUERY_COMMAND)
HELP_SEARCH_STR = '!{} <search terms>'.format(SEARCH_COMMAND)
HELP_REDIRECT_STR = '!{} <redirect from>:<redirect to>'.format(REDIRECT_COMMAND)
HELP_REMOVE_REDIRECT_STR = '!{} <entry>'.format(REMOVE_REDIRECT_COMMAND)
HELP_WHOWROTE_STR = '!{} <entry> [: num]'.format(WHOWROTE_COMMAND)

# TODO: Clean all of this up.
DOCS_STR = (
    'To define a glossary entry: `{}`. '
    'To get a definition: `{}`. '
    'To search for entries: `{}`. '
).format(HELP_DEFINE_STR, HELP_QUERY_STR, HELP_SEARCH_STR)

OOPS_BASE = "I didn't understand that. "
OOPS_GENERIC = OOPS_BASE + DOCS_STR
OOPS_REDIRECT = OOPS_BASE + 'Try ' + HELP_REDIRECT_STR
OOPS_SEARCH = OOPS_BASE + 'Try ' + HELP_SEARCH_STR

ADD_DEFINITION_RESULT_TEMPLATE = (
    u'Okay! "{entry}" is now "{definition}". '
    u'This is the {nth} time it has been defined.'
)

QUERY_RESULT_TEMPLATE = (
    u'{entry} ({num}/{total}): {definition} [{age}]'
)

UNDEFINED_TEMPLATE = u'"{}" is undefined.'

INVALID_ENTRY_CHARS = [c for c in string.punctuation if c not in ['_', '-']]


class InvalidEntryError(Exception):
    pass


class InvalidRedirectError(Exception):
    pass


class InvalidEntryNumberError(Exception):
    """
    Raised when user requests a numbered entry that does not exist.
    """
    def __init__(self, entry, entries, redirect=None):
        count = len(entries)
        valid_records_str = 'Valid record numbers are 1-{}.'.format(count)

        if redirect:
            message = (
                u'"{}" redirects to "{}," which has {} records. {}'.format(
                    entry, redirect.entry, count, valid_records_str
                )
            )
        else:
            message = (
                u'"{}" has {} records. {}'.format(
                    entry, count, valid_records_str
                )
            )

        super(InvalidEntryNumberError, self).__init__(message)


def override_command(name, aliases=None, doc=None, priority=2):
    """
    Command decorator that accepts a priority argument.

    This mimics the standard ``command`` decorator, except it accepts
    a priority int that is passed through to the CommandHandler
    instance.

    As documented in pmxbot.core.Handler, higher value == higher priority.
    """
    handler = CommandHandler(priority=priority, name=name.lower(), doc=doc)

    if aliases:
        aliases = [
            AliasHandler(name=alias, parent=handler) for alias in aliases
        ]

        handler.aliases = aliases

    return handler.decorate


class Glossary(storage.SelectableStorage):
    """
    Glossary class.

    The usage of SelectableStorage, SQLiteStorage, and cls.store are cribbed
    from the pmxbot quotes module.
    """
    @classmethod
    def initialize(cls, db_uri=None, load_fixtures=True):
        db_uri = db_uri or pmxbot.config.database
        cls.store = cls.from_URI(db_uri)

        if load_fixtures:
            cls.load_fixtures()

        cls._finalizers.append(cls.finalize)

    @classmethod
    def finalize(cls):
        del cls.store

    @classmethod
    def load_fixtures(cls, path=None):
        config_path_key = 'glossary_fixtures_path'

        if not path:
            path = pmxbot.config.get(config_path_key)

        if not path:
            print('- No fixtures path provided.')
            return

        try:
            with open(path) as f:
                print('- Loading fixtures from ' + path)
                data = json.load(f)
                cls.save_entries(data)
        except IOError:
            print('- No fixtures file found at path {}'.format(path))

    @classmethod
    def save_entries(cls, data):
        """
        Save a dictionary of entries and definitions to the store.

        If the definition already exists in the history for the entry,
        this will not re-add it.
        """
        for entry, definition in data.items():
            existing = cls.store.get_all_records_for_entry(entry)
            existing_defs = [e.definition for e in existing]

            if definition not in existing_defs:
                cls.store.add_entry(entry, definition, 'the defaults')


class SQLiteGlossary(Glossary, storage.SQLiteStorage):
    CREATE_GLOSSARY_SQL = """
      CREATE TABLE IF NOT EXISTS glossary (
       entryid INTEGER PRIMARY KEY AUTOINCREMENT,
       entry VARCHAR NOT NULL,
       entry_lower VARCHAR NOT NULL,
       definition TEXT NOT NULL,
       author VARCHAR NOT NULL,
       channel VARCHAR,
       timestamp DATE DEFAULT (datetime('now','utc'))
    )
    """

    CREATE_GLOSSARY_INDEX_SQL = """
      CREATE INDEX IF NOT EXISTS ix_glossary_entry ON glossary(entry_lower)
    """

    CREATE_REDIRECTS_SQL = """
      CREATE TABLE IF NOT EXISTS glossary_redirects (
       redirectid INTEGER PRIMARY KEY AUTOINCREMENT,
       redirect_from VARCHAR UNIQUE NOT NULL,
       redirect_to VARCHAR NOT NULL
    )
    """

    CREATE_REDIRECT_INDEX_SQL = """
      CREATE INDEX IF NOT EXISTS ix_glossary_redirect
      ON glossary_redirects(redirect_from)
    """

    ALL_ENTRIES_CACHE_KEY = 'all_entries'

    cache = {}

    @staticmethod
    def date_str_to_datetime(date_str):
        """
        Helper function to turn a queried date string into a datetime object.

        Simple, but used in multiple places.
        """
        return datetime.datetime.utcfromtimestamp(int(date_str))

    def init_tables(self):
        self.db.execute(self.CREATE_GLOSSARY_SQL)
        self.db.execute(self.CREATE_GLOSSARY_INDEX_SQL)
        self.db.execute(self.CREATE_REDIRECTS_SQL)
        self.db.execute(self.CREATE_REDIRECT_INDEX_SQL)
        self.db.commit()

    def dump_to_json(self):
        """
        Dumps all entry data to a temporary file.
        """
        entries, redirects = [], []

        outfile = tempfile.NamedTemporaryFile(
            mode='w',
            prefix='pmxbot-glossary_dump',
            suffix='.json',
            delete=False
        )

        entry_sql = """
          SELECT entry,
            entry_lower,
            definition,
            author,
            channel,
            strftime('%s', timestamp)
          FROM Glossary
          ORDER BY entry_lower
        """

        for row in self.db.execute(entry_sql):
            entries.append({
                'entry': row[0],
                'entry_lower': row[1],
                'definition': row[2],
                'author': row[3],
                'channel': row[4],
                'timestamp': row[5]
            })

        redirect_sql = """
          SELECT redirect_from, redirect_to
          FROM glossary_redirects
        """

        for row in self.db.execute(redirect_sql):
            redirects.append({
                'redirect_from': row[0],
                'redirect_to': row[1],
            })

        dump_data = {
            'entries': entries,
            'redirects': redirects,
        }

        json.dump(dump_data, outfile, indent=2)

        outfile.close()

        return dump_data, outfile.name

    def load_from_json(self, filepath):
        """
        Load entries from json data in filepath (str).

        The file should contain data dumped by ``dump_to_json``. If you're
        trying to create fixtures, use a fixtures file handled by
        ``initialize``.
        """
        records_by_entry_lower = {}

        inserted = []

        with open(filepath, 'r') as f:
            data = json.load(f)

            for entry_data in data.get('entries', []):
                entry = entry_data['entry']
                entry_lower = entry_data['entry_lower']
                definition = entry_data['definition']
                timestamp = entry_data['timestamp']
                datetime_ = self.date_str_to_datetime(timestamp)

                if not entry_lower in records_by_entry_lower:
                    records_by_entry_lower[entry_lower] = (
                        self.get_all_records_for_entry(entry_lower)
                    )

                existing_records = records_by_entry_lower[entry_lower]

                already_exists = False

                for existing in existing_records:
                    if all([
                        existing.entry == entry,
                        existing.definition == definition,
                        existing.datetime == datetime_
                    ]):
                        already_exists = True
                        break

                if not already_exists:
                    self.add_entry(
                        entry,
                        definition,
                        entry_data['author'],
                        entry_data['channel'],
                        datetime_
                    )

                    inserted.append(entry_data)

            for redirect_data in data.get('redirects', []):
                self.add_redirect(
                    redirect_data['redirect_from'],
                    redirect_data['redirect_to']
                )

            return data, inserted

    def bust_all_entries_cache(self):
        if self.ALL_ENTRIES_CACHE_KEY in self.cache:
            del self.cache[self.ALL_ENTRIES_CACHE_KEY]

    def add_redirect(
        self,
        redirect_from,
        redirect_to
    ):
        """
        Add redirection from one entry to another.

        ``redirect_from`` does not need to exist in the glossary table.
        """
        redirect_from = redirect_from.lower()
        redirect_to = redirect_to.lower()

        if not self.get_latest_record(redirect_to):
            raise InvalidRedirectError(
                u'"{}" is not defined'.format(redirect_to)
            )

        existing = self.get_redirect(redirect_to)

        if existing:
            raise InvalidRedirectError(
                u'"{}" is itself being redirected to "{}."'.format(
                    redirect_to, existing.entry
                )
            )

        sql = """
            INSERT OR REPLACE INTO glossary_redirects
              (redirectid, redirect_from, redirect_to)
            VALUES (
              (
                SELECT redirectid from glossary_redirects
                WHERE redirect_from = ?
              ),
              ?,
              ?
            )
        """

        self.db.execute(sql, (redirect_from, redirect_from, redirect_to))
        self.db.commit()

    def remove_redirect(self, entry):
        sql = """
          DELETE FROM glossary_redirects
          WHERE redirect_from = ?
        """

        self.db.execute(sql, (entry.lower(), ))
        self.db.commit()

    def get_redirect(self, entry):
        sql = """
            SELECT redirect_to
            FROM glossary_redirects
            WHERE redirect_from = ?
        """

        row = self.db.execute(sql, (entry.lower(), )).fetchone()

        if row:
            return self.get_latest_record(row[0])

        return None

    def add_entry(
        self,
        entry,
        definition,
        author,
        channel=None,
        timestamp=None
    ):
        redirect_entry = self.get_redirect(entry)

        if redirect_entry:
            msg = (
                u'"{}" redirects to "{}." Redirected entries cannot be defined.'
            ).format(entry, redirect_entry.entry)

            raise InvalidEntryError(msg)

        if timestamp:
            sql = """
              INSERT INTO glossary
                (entry, entry_lower, definition, author, channel, timestamp)
              VALUES (?, ?, ?, ?, ?, ?)
            """
            values = (
                entry, entry.lower(), definition, author, channel, timestamp
            )
        else:
            sql = """
              INSERT INTO glossary
                (entry, entry_lower, definition, author, channel)
              VALUES (?, ?, ?, ?, ?)
            """
            values = (entry, entry.lower(), definition, author, channel)

        self.db.execute(sql, values)
        self.db.commit()
        self.bust_all_entries_cache()

        return self.get_latest_record(entry)

    def get_all_records(self):
        """
        Returns list of all the latest entries in the glossary.
        """
        cache_key = 'all_entries'
        entries = self.cache.get(cache_key)

        if entries is None:
            sql = """
              SELECT entry,
                entry_lower,
                definition,
                author,
                channel,
                strftime('%s', timestamp),
                COUNT(entry_lower)
              FROM glossary
              GROUP BY entry_lower
              ORDER BY timestamp
            """

            query = self.db.execute(sql).fetchall()
            entries = []

            for row in query:
                count = row[-1]
                record = GlossaryRecord(
                    row[0],
                    row[1],
                    row[2],
                    row[3],
                    row[4],
                    datetime.datetime.utcfromtimestamp(int(row[5])),
                    count - 1,
                    count
                )

                entries.append(record)

            entries.sort(key=lambda x: x.entry_lower)

            self.cache[cache_key] = entries

        return entries

    def get_random_entry(self):
        """
        Returns a random entry from the glossary.
        """
        entries = self.get_all_records()

        if not entries:
            return None

        return random.choice(entries).entry

    def get_latest_record(self, entry):
        """
        Returns GlossaryQueryResult for an entry in the glossary.
        """
        records = self.get_all_records_for_entry(entry)

        if not records:
            return None

        return records[-1]

    def get_nth_record(self, entry, num, follow_redirect=True):
        """
        Returns GlossaryQueryResult for nth entry in the glossary.
        """
        redirect = Glossary.store.get_redirect(entry)
        target_entry = redirect.entry if redirect else entry

        try:
            query_result = Glossary.store.get_entry_data(target_entry, num)
        except IndexError:
            entries = Glossary.store.get_all_records_for_entry(target_entry)

            if not entries:
                return UNDEFINED_TEMPLATE.format(entry)

            raise InvalidEntryNumberError(entry, entries, redirect)


        # Entry numbering starts at 1
        if num is not None and num < 1:
            raise IndexError

        stored = self.get_all_records_for_entry(entry)

        if not stored:
            return None

        if num:
            return stored[num - 1]

        return stored[-1]

    def get_all_records_for_entry(self, entry):
        """
        Returns a list of objects for all definitions of an entry.
        """
        entry = entry.lower()

        sql = """
            SELECT entry,
              entry_lower,
              definition,
              author,
              channel,
              strftime('%s', timestamp)
            FROM glossary
            WHERE entry_lower LIKE ?
            ORDER BY timestamp
        """

        results = self.db.execute(sql, (entry, )).fetchall()

        entry_data = []
        total_count = len(results)

        for i, row in enumerate(results):
            entry_data.append(
                GlossaryRecord(
                    row[0],
                    row[1],
                    row[2],
                    row[3],
                    row[4],
                    self.date_str_to_datetime(row[5]),
                    i,
                    total_count
                )
            )

        return entry_data

    def get_similar_words(self, search_str):
        search_str = search_str.lower()
        all_entries = self.get_all_records()

        matches = [e.entry for e in all_entries if search_str in e.entry_lower]

        return matches

    def search_definitions(self, search_str):
        """
        Returns entries whose definitions contain the search string.
        """
        search_str = '%{}%'.format(search_str)

        sql = """
            SELECT DISTINCT entry
            FROM glossary
            WHERE definition LIKE ?
            ORDER BY entry
        """

        results = self.db.execute(sql, (search_str, ))

        return [r[0] for r in results]


GlossaryRecord = namedtuple(
    'GlossaryQueryResult',
    'entry entry_lower definition author channel datetime index total_count'
)


class QueryHandler(object):
    RESPONSE_TEMPLATE = (
        u'{entry} ({num}/{total}): {definition} [{age}]'
    )

    def __init__(self, entry, num=None):
        self.entry = entry
        self.num = num
        self.redirect = Glossary.store.get_redirect(entry)

        if self.redirect:
            self.target_entry = self.redirect.entry
        else:
            self.target_entry = self.entry

        self.records = Glossary.store.get_all_records_for_entry(
            self.target_entry
        )

        self.success = bool(self.records) and self.num_is_valid

    def response(self):
        if not self.success:
            return self.error_response()

        target_record = self.target_record

        response = self.RESPONSE_TEMPLATE.format(
            entry=target_record.entry,
            num=target_record.index + 1,
            total=target_record.total_count,
            definition=target_record.definition,
            age=datetime_to_age_str(target_record.datetime)
        )

        if self.redirect:
            response = u'{} redirects to {}'.format(self.entry, response)

        return response

    def error_response(self):
        if self.success:
            return None

        message = None

        if not self.records:
            message = u'"{}" is undefined'.format(self.entry)
            # Check if there are any similar entries that may be relevant.
            suggestions = list(get_alternative_suggestions(self.entry))[:10]

            if suggestions:
                message += (
                    u'. May I interest you in {}?'.format(readable_join(suggestions))
                )

        elif not self.num_is_valid:
            message = unicode(
                InvalidEntryNumberError(
                    self.entry, self.records, self.redirect
                )
            )

        return message or 'Something strange happened.'

    @property
    def target_record(self):
        if self.num:
            if not self.num_is_valid:
                raise InvalidEntryNumberError(
                    self.entry, self.records, self.redirect
                )

            return self.records[self.num - 1]

        return self.records[-1]

    @property
    def num_is_valid(self):
        if self.num is None:
            return True

        return 1 <= self.num <= len(self.records)


class WhoWroteHandler(QueryHandler):
    def response(self):
        if not self.success:
            return self.error_response()

        response = u'{} authored the {} definition of {}.'.format(
            self.target_record.author,
            nth_str(self.target_record.index + 1),
            self.target_record.entry
        )

        if self.redirect:
            response = u'{} redirects to {}. {}'.format(
                self.entry, self.target_record.entry, response
            )

        return response


def datetime_to_age_str(dt):
    """
    Returns a human-readable age given a datetime object.
    """
    age = (datetime.datetime.utcnow() - dt)
    days = age.days

    if days >= 365:
        return '{0:.1f} years ago'.format(days / 365.0)

    if days > 30:
        return '{0:.1f} months ago'.format(days / 30.5)

    if days > 1:
        return '{} days ago'.format(days)

    if days == 1:
        return 'yesterday'

    seconds = age.total_seconds()
    hours = int(seconds / 3600)

    if hours > 1:
        return '{} hours ago'.format(hours)

    if hours == 1:
        return '1 hour ago'

    minutes = int(seconds / 60)

    if minutes > 1:
        return '{} minutes ago'.format(minutes)

    if minutes == 1:
        return '1 minute ago'

    return 'just now'


def readable_join(items, conjunction='or'):
    """
    Returns an oxford-comma-joined string with "or."

    E.g., ['thing1', 'thing2', 'thing3'] becomes "thing1, thing2, or thing3".
    """
    if not items:
        return None

    count = len(items)

    if count == 1:
        s = items[0]
    elif count == 2:
        template = u' {} '.format(conjunction)
        s = template.join(items)
    else:
        s = u'{}, {} {}'.format(u', '.join(items[:-1]), conjunction, items[-1])

    return s


def get_alternative_suggestions(entry):
    """
    Returns a set of entries that may be similar to the provided entry.

    Useful for trying to find near misses. E.g., "run" is not defined but
    "running" is.
    """
    query_words = set()

    for delim in (' ', '-', '_'):
        for part in entry.split(delim):
            query_words.add(part.strip().lower())

    results = set()

    for word in query_words:
        similar = Glossary.store.get_similar_words(word)

        if similar:
            results |= set(similar)

    return results


def handle_nth_definition(entry, num=None):
    """
    Returns a formatted result string for an entry.

    If ``num`` is passed, it will return the corresponding numbered, historical
    definition for the entry.
    """
    redirect = Glossary.store.get_redirect(entry)
    target_entry = redirect.entry if redirect else entry

    try:
        query_result = Glossary.store.get_entry_data(target_entry, num)
    except IndexError:
        entries = Glossary.store.get_all_records_for_entry(target_entry)

        if not entries:
            return UNDEFINED_TEMPLATE.format(entry)

        raise InvalidEntryNumberError(entry, entries, redirect)

    if query_result:
        kwargs = dict(
            entry=query_result.entry,
            num=query_result.index + 1,
            total=query_result.total_count,
            definition=query_result.definition,
            age=datetime_to_age_str(query_result.datetime),
        )

        response = QUERY_RESULT_TEMPLATE.format(**kwargs)

        if redirect:
            response = u'{} redirects to {}'.format(entry, response)

        return response

    response = UNDEFINED_TEMPLATE.format(entry)

    # Check if there are any similar entries that may be relevant.
    suggestions = list(get_alternative_suggestions(entry))[:10]

    if suggestions:
        response += (
            u' May I interest you in {}?'.format(readable_join(suggestions))
        )

    return response


def entry_number_command(accepts_num, docs, require_entry=False):
    """
    Decorator for commands that care about an entry string and possibly an
    entry number.
    """
    def func_wrapper(func):
        def inner(client, event, channel, nick, rest):
            rest = rest.strip()
            num = None

            if rest.lower() == 'help':
                return docs

            if ':' in rest:
                if not accepts_num:
                    return docs

                parts = rest.split(':', 1)
                entry, num = parts[0].strip(), parts[1].strip()

                try:
                    num = int(num)
                except (ValueError, TypeError):
                    return docs
            else:
                entry = rest

            if require_entry and not entry:
                return docs

            if accepts_num:
                return func(entry, num)

            return func(entry)

        return inner
    return func_wrapper


def nth_str(num):
    num = int(num)

    nth_map = {
        1: '1st',
        2: '2nd',
        3: '3rd',
    }

    return nth_map.get(num, '{}th'.format(num))


@override_command(
    DEFINE_COMMAND, doc=HELP_DEFINE_STR, aliases=('set', 'gdefine')
)
def define(client, event, channel, nick, rest):
    """
    Add a definition for a glossary entry.
    """
    rest = rest.strip()

    if rest.lower() == 'help':
        return DOCS_STR

    if not ':' in rest:
        return OOPS_GENERIC

    if '::' in rest:
        return "I can't handle '::' right now. Please try again without it."

    parts = rest.split(':', 1)

    if len(parts) != 2:
        return OOPS_GENERIC

    entry = parts[0].strip()

    invalid_char = next((c for c in entry if c in INVALID_ENTRY_CHARS), None)

    if invalid_char:
        return '"{}" cannot be used in a glossary entry.'.format(
            invalid_char
        )

    definition = parts[1].strip()

    existing = Glossary.store.get_latest_record(entry)

    if existing and existing.definition == definition:
        return "That's already the current definition."

    try:
        result = Glossary.store.add_entry(entry, definition, nick, channel)
    except InvalidEntryError as e:
        return str(e)

    return ADD_DEFINITION_RESULT_TEMPLATE.format(
        entry=result.entry,
        definition=result.definition,
        nth=nth_str(result.total_count)
    )


@command(QUERY_COMMAND, doc=DOCS_STR)
@entry_number_command(accepts_num=True, docs=HELP_QUERY_STR)
def query_command(entry, num):
    """
    Retrieve a definition of an entry.
    """
    if not entry:
        entry = Glossary.store.get_random_entry()
        num = None

    return QueryHandler(entry, num).response()


@command(REDIRECT_COMMAND)
def redirect_command(client, event, channel, nick, rest):
    """
    Redirect one entry to another.
    """
    if ':' not in rest:
        return OOPS_REDIRECT

    parts = rest.split(':')

    if len(parts) != 2:
        return OOPS_REDIRECT

    redirect_from, redirect_to = [p.strip() for p in parts]

    try:
        Glossary.store.add_redirect(redirect_from, redirect_to)
    except InvalidRedirectError as e:
        return str(e)

    return u'"{}" will now redirect to "{}"'.format(redirect_from, redirect_to)


@command(REMOVE_REDIRECT_COMMAND)
@entry_number_command(
    accepts_num=False, require_entry=True, docs=HELP_REMOVE_REDIRECT_STR
)
def remove_redirect(entry):
    """
    Remove a redirect.
    """
    existing = Glossary.store.get_redirect(entry)

    if not existing:
        return u'"{}" is not being redirected anywhere.'

    Glossary.store.remove_redirect(entry)

    return u'"{}" is no longer being redirected to "{}"'.format(
        entry, existing.entry
    )


@command(SEARCH_COMMAND, doc=DOCS_STR)
@entry_number_command(
    accepts_num=False, require_entry=True, docs=HELP_SEARCH_STR
)
def search(entry):
    """
    Search the entries and defintions.
    """
    entry_matches = set(Glossary.store.get_similar_words(entry))
    lower_matches = {e.lower() for e in entry_matches}

    def_matches = set(
        e for e in Glossary.store.search_definitions(entry)
        if e.lower() not in lower_matches
    )

    matches = sorted(list(entry_matches | def_matches))

    if not matches:
        return 'No glossary results found.'
    else:
        result = (
            u'Found glossary entries: {}. To get a definition: `!{} <entry>`'
        ).format(readable_join(matches, conjunction='and'), QUERY_COMMAND)

        return result


@command(WHOWROTE_COMMAND, doc=HELP_WHOWROTE_STR)
@entry_number_command(
    accepts_num=True, require_entry=True, docs=HELP_WHOWROTE_STR
)
def who_wrote(entry, num=None):
    """
    Search the entries and defintions.
    """
    return WhoWroteHandler(entry, num).response()


def archives_link(rest, num=None, url_base=None):
    """
    Coming soon.
    """
    slack_url = url_base or pmxbot.config.get('slack_url')

    if not slack_url:
        return 'Slack URL is not configured.'

    entry = Glossary.store.get(rest, num)

    if not entry:
        return UNDEFINED_TEMPLATE.format(entry)

    timestamp = calendar.timegm(entry.datetime.utctimetuple())

    channel = entry.channel

    if not channel:
        return (
            u"I don't know which channel {} was defined in. "
            u"If it's redefined, I can try again."
        ).format(rest)

    if channel.startswith('#'):
        channel = channel[1:]

    url = '{slack_url}/archives/{channel}/p{timestamp}'.format(
        slack_url=slack_url,
        channel=channel,
        timestamp=timestamp * 1000000
    )

    template = '[Experimental] Attempting to link to Slack archives at {}: {}'

    return template.format(entry.datetime.ctime(), url)


@command(ADMIN_COMMAND)
def admin(client, event, channel, nick, rest):
    rest = rest.strip()

    version = pmx_glossary.__version__
    version_str = 'Using pmxbot-glossary v{}'.format(version)

    if rest == 'status':
        # TODO: Version, entry count, etc.
        pass

    if rest == 'version':
        return version_str

    if rest == 'repo':
        return 'https://github.com/harveyr/pmxbot-glossary/'

    if rest == 'changelog':
        return (
            'https://github.com'
            '/harveyr/pmxbot-glossary/blob/master/changelog.md'
            ' (Currently using v{})'
        ) .format(version)

    return OOPS_BASE



#!/usr/bin/env python
from setuptools import setup, find_packages

import pmx_glossary

setup(
    name='pmxbot-glossary',
    version=pmx_glossary.__version__,
    description='A pmxbot glossary extension.',
    author='Harvey Rogers',
    author_email='harveyr@gmail.com ',
    url='https://github.com/harveyr/pmxbot-glossary/',
    license='MIT',
    packages=find_packages(),
    install_requires=[
        'pmxbot',
        'python-dateutil',
        'click',
    ],
    entry_points=dict(
        console_scripts=[
            'pmxglos = pmx_glossary.cli:cli'
        ],
        pmxbot_handlers=[
            'Glossary = pmx_glossary.glossary:Glossary.initialize',
        ]
    ),
)
